/*
	midiout-test - Test application for MIDI out interface
	http://gba.gabiot.com
	Copyright (C) 2006  Avelino Herrera Morales

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "gba.h"
#include "font_8x8.h"

#define  PRESSED(k)  (!(REG_KEYPAD & (k)))

int __gba_multiboot;

/* modo grfico */
#define BG0_ENABLE      0x100
#define BG1_ENABLE      0x200
#define BG2_ENABLE      0x400
#define BG3_ENABLE      0x800
#define OBJ_ENABLE      0x1000
#define WIN0_ENABLE     0x2000
#define WIN1_ENABLE     0x4000
#define WINOBJ_ENABLE   0x8000

#define BACK_BUFFER     0x10

#define MODE0           0x0
#define MODE1           0x1
#define MODE2           0x2
#define MODE3           0x3
#define MODE4           0x4
#define MODE5           0x5

#define  VIDEO_BUFFER_0  ((u16 *) 0x6000000)
#define  VIDEO_BUFFER_1  ((u16 *) 0x600A000)
#define  BG_PALETTE_MEM  ((u16 *) 0x5000000)

#define  TOP_LEFT      0
#define  TOP_RIGHT     239
#define  BOTTOM_LEFT   38060
#define  BOTTOM_RIGHT  38299

#define RGB16(r,g,b)  ((r) + ((g) << 5) + ((b) << 10))

#define  ERASE_SCREEN(scr)  memset(scr, 0, 240 * 160)


u16 *current_video_buffer = VIDEO_BUFFER_0;

struct {
	u8 *data;
	u8 data_size;
	s8 next_bit;
	u8 byte;
} midi_send_struct;

// SO = 3.3 v   (always 1)  (SO is SI in the slave end of the cable)
// SC = MIDI stream         (SC is SC too in the slave end of the cable)
#define  REG_R_0_VALUE  0x80F8
#define  REG_R_1_VALUE  0x80F9

void InterruptProcess(void) __attribute__ ((section(".iwram")));   // the ISR must be declared in this way

 
void InterruptProcess(void) {
	u8 bit;

	if (midi_send_struct.byte >= midi_send_struct.data_size)
		bit = 1;
	else if (midi_send_struct.next_bit == -1) {
		midi_send_struct.next_bit = 0;
		bit = 0;
	}
	else if (midi_send_struct.next_bit > 7) {
		midi_send_struct.byte++;
		midi_send_struct.next_bit = -1;
		bit = 1;
	}
	else {
		bit = 1 & (midi_send_struct.data[midi_send_struct.byte] >> midi_send_struct.next_bit);
		midi_send_struct.next_bit++;
	}
	REG_R = (bit) ? REG_R_1_VALUE : REG_R_0_VALUE;
	// clear interrupt flag
	REG_IF |= REG_IF;
}


void midi_start() {
	midi_send_struct.data_size = 0;
	midi_send_struct.byte = -1;
	// stop timer 0
	REG_TM0CNT_H = 0;
	// enable irq for timer 0
	REG_IE = 0x0008;
	REG_IME = 1;
	// load timer 0
	// 65536 - 64999 = 537
	// (16 * 1024 * 1024) / 537 = 31242 bps
	// MIDI is 31250 bsp at +/-1% tolerance (-1% => 30937, +1% => 31562)
	// so 31242 bps is a valid bps value
	REG_TM0CNT_L = 0xFDE7;   // 64999
	REG_TM0CNT_H = 0x80 | 0x40;   // start + irq + f (16MHz)
}


void midi_stop() {
	// stop timer 0
	REG_TM0CNT_H = 0;
}


u8 midi_send(u8 *data, u8 size) {
	if (midi_send_struct.byte < midi_send_struct.data_size)
		return 0;
	midi_send_struct.data = data;
	midi_send_struct.data_size = size;
	midi_send_struct.next_bit = -1;
	midi_send_struct.byte = 0;
	return 1;
}


void flip_and_show(void) {
	if (current_video_buffer == VIDEO_BUFFER_1) {
		REG_DISPCNT |= BACK_BUFFER;
		current_video_buffer = VIDEO_BUFFER_0;
	}
	else {
		REG_DISPCNT &= ~BACK_BUFFER;
		current_video_buffer = VIDEO_BUFFER_1;
	}
}


void put_character(char c, int x, int y, u16 *video_buffer, int erase) {
	int font_y = (((int) c) - 32) / 16;
	int font_x = (((int) c) - 32) % 16;
	int i, j;
	int font_offset = (font_y * 64 * 8) + (font_x * 4);
	int offset = ((y * 240) + x) >> 1;
	for (i = 0; i < 8; i++, offset += 120, font_offset += 64) {
		if (erase) {
			for (j = 0; j < 4; j++)
				video_buffer[offset + j] = 0x0000;
		}
		for (j = 0; j < 4; j++) {
			register u16 d = fontData[font_offset + j];
			register u16 prev = video_buffer[offset + j];
			register u16 to_write = 0;
			if ((d & 0x00FF) != fontTransparentIndex)
				to_write |= (d & 0x00FF);
			else
				to_write |= (prev & 0x00FF);
			if ((d >> 8) != fontTransparentIndex)
				to_write |= (d & 0xFF00);
			else
				to_write |= (prev & 0xFF00);
			video_buffer[offset + j] = to_write;
		}
	}
}


void put_string(char *s, int x, int y, u16 *video_buffer, int erase) {
	int i = 0;
	while (s[i] != 0) {
		put_character(s[i], x, y, video_buffer, erase);
		x += 8;
		i++;
	}
}


void wait_for_vsync() {
	while (((volatile u16) REG_VCOUNT) != 160)
		;
}


#define  GET_X_CENTER(width)  (120 - ((width) * 4))


void show_main_screen() {
	ERASE_SCREEN(current_video_buffer);
	put_string("MIDI Out tester", GET_X_CENTER(15), 30, current_video_buffer, 0);
	put_string("A: note on ", GET_X_CENTER(11), 50, current_video_buffer, 0);
	put_string("B: note off", GET_X_CENTER(11), 60, current_video_buffer, 0);
	wait_for_vsync();
	flip_and_show();
}


int AgbMain(void) {
	int i;
	u8 note_on_message[] = {0x90, 0x3C, 0x40};
	u8 note_off_message[] = {0x90, 0x3C, 0x00};

	REG_DISPCNT = MODE4 | BG2_ENABLE;
	for (i = 0; i < 256; i++)
		BG_PALETTE_MEM[i] = fontPalette[i];
	midi_start();
	while (1) {
		show_main_screen();
		if (PRESSED(BUTTON_A)) {
			midi_send(note_on_message, 3);
			while (PRESSED(BUTTON_A))
				;
		}
		else if (PRESSED(BUTTON_B)) {
			midi_send(note_off_message, 3);
			while (PRESSED(BUTTON_B))
				;
		}
	}
}

